/** @file TextureManager.cpp
*/
#include <assert.h>
#include <sps2lib.h>
#include <sps2tags.h>
#include <memory.h>

#include "PS2Utils.h"
#include "TextureManager.h"
 
#define DEBUG
/** .
	 This is a simple routine because all swizzling, converting, resizing, etc etc
	 is done previously by texture converter program. This way of proceeding is 
	 often used, to reduce run time complexity of the program.
	 In this first release only 32 bit clut entries are supported, class must be 
	 extended to permit the use of 16 bit clut format.
	 To correctly read the header, first the header structure must be read, then
	 seek the file at first byte after the header, and header size must be taken
	 by structure itself.
	 @param FileName Name of the file that contain texture, file must complain 
						  structure of the file generated by Texture Converter software
    @param SPS2Descriptor Descriptor of SPS2 library, used to manage memory.
*/
NablaPS2Utils::TextureManager::TextureManager(const char *FileName, int SPS2Descriptor) {

	//Do not forget to initialize all Class internal members
	m_status			= false;
	m_imageData		= NULL;
	m_clutData		= NULL;

	//Open the file
	FILE * fp = NULL;
	fp = fopen(FileName, "rb");
	if (fp) {
		
		// Read all the header, check that reading went good.
		TextureFileHeader TFH;
		int Read = fread(&TFH, sizeof(TFH), 1, fp);
		assert(1 == Read);
		UpdateImageData(TFH);

		//Now read all data in a while, remember that TFH contains datasize of the 
		//binary data of image.
		m_SPS2Desc		= SPS2Descriptor;
		m_numMemBlocks = (TFH.DataSize + 4095) / 4096; 
		m_imageData		= sps2Allocate(4096 * m_numMemBlocks, 
									SPS2_MAP_BLOCK_4K | SPS2_MAP_UNCACHED,	m_SPS2Desc);
		fread(m_imageData->pvStart, 1, TFH.DataSize, fp);
#ifdef DEBUG
		printf("Image size = %d Bytes, Num of 4096 block = %d\n", TFH.DataSize, m_numMemBlocks); 
#endif
		//Now if a CLUT is present allocate memory for the clut, 
		if (PSM_PSMT8 == TFH.PixelFormat) {
		
			//256 CLUT entry are present, so 256 * 2 = 1024 bytes are needed
			m_clutData = sps2Allocate(4096, 
									SPS2_MAP_BLOCK_4K | SPS2_MAP_UNCACHED,	m_SPS2Desc);
			fread(m_clutData->pvStart, 1, 1024, fp);
		} 
		else if (PSM_PSMT4 == TFH.PixelFormat) {
		
			//16 CLUT entry are present, so 16 * 4 = 64 byte of memory should be 
			//allocated
			m_clutData = sps2Allocate(4096, 
									SPS2_MAP_BLOCK_4K | SPS2_MAP_UNCACHED,	m_SPS2Desc);
			fread(m_clutData, 1, 64, fp);
		}

		//All went good.
		m_status = true;
	}
}

/** .
	 Simply release allocated memory for the texture.
*/
NablaPS2Utils::TextureManager::~TextureManager() {

	SPS2_SAFE_FREE(m_imageData);
	SPS2_SAFE_FREE(m_clutData);
}

/** .
	 DMA packet is build sequentially updating pointer to memory that hold the
	 packet. This routine assume that texture data is to be uploaded in the first
	 location of GSAdress, remember that is possible to transfer texture not only 
	 at the beginning of the buffer using TRXPOS register.
	 @param GSAddress Address in GS where to upload texture into, remember that
							this address must be in word address / 64, so since a word
							is 4 bytes long, to translate from byte address to word / 64
							address it is necessary to divide by 256. If original address
							is in page address, since a page is 8KB, it is necessary 
							multiplying address by 32 since 8196 / 256 = 32.
	 @param DMAMemory Memory used to build DMA packet. Since dma packet is small and
							is not possible to allocate memory in chunk smaller than 4KB
							it is convenient allocate a single 4KB area in main program and
							pass this buffer to every routine that need to do DMA transfer.
*/
void NablaPS2Utils::TextureManager::SendTexture(
										sps2uint32 GSAddress, 
										sps2Memory_t *DMAMemory) {

	//! Update upload address.
	m_UploadAddress = GSAddress;

	//To upload a texture 4 register must be set, moreover GIFTag for begin 
	//IMAGE MODE transfer must also be set, this is a total of 
	//GifTag to specify 4 register + 4 * register specs + GifTag for IMAGEMODE
	//equal 6 qword of data.
	PS2_QWORD *DMAPacket = (PS2_QWORD *) DMAMemory->pvStart;
	DMAPacket++->ul128 = PS2_DMA_SET_HEADER(6, 0, DMA_ID_SOURCE_CNT, 0, 0, 0);

	//Now giftag, A+D mode and set other registers. Remember that EOP paramter 
	//Must be set to GS_GIF_EOP_WITHFOLLOWINGPRIM because we must send three
	//GS Primitives.
	sceGifTag *GifTag = (sceGifTag *) DMAPacket;
	memset(GifTag, 0, sizeof(sceGifTag));
	DMAPacket++->ul64[0] = PS2_GIF_SET_TAG(4, 
								  GS_GIF_EOP_WITHFOLLOWINGPRIM, 
								  GS_GIF_PRE_IGNOREPRIM, 
								  0, 
								  GIF_FLG_PACKED, 
								  1);
	GifTag->REGS0 = GIF_REG_A_D;

	//(m_width + 63) / 64 gives smaller buffer width for transfer.
	int BufferWidth = (m_width + 63) / 64;
	DMAPacket->ul64[0] = PS2_GIF_SET_BITBLTBUF(0, 0, 0, GSAddress, BufferWidth, m_pixelFormat);
	DMAPacket++->ul64[1] = GIF_A_D_REG_BITBLTBUF;
	DMAPacket->ul64[0] = PS2_GIF_SET_TRXPOS(0, 0, 0, 0, 0);
	DMAPacket++->ul64[1] = GIF_A_D_REG_TRXPOS;
	DMAPacket->ul64[0] = PS2_GIF_SET_TRXREG(m_width, m_height);
	DMAPacket++->ul64[1] = GIF_A_D_REG_TRXREG;
	//Start image transfer.
	DMAPacket->ul64[0] = PS2_GIF_SET_TRXDIR(0);
	DMAPacket++->ul64[1] = GIF_A_D_REG_TRXDIR;

	//Now another GIFTag specifiyng IMAGE transfer is to be stored into first
	//cnt DMA tag packet, this GIFtag ends data transferred by the first chain
	//tag, all the data of this GIF packet are to be sent with a series of
	//ref chain tag.
	GifTag = (sceGifTag *) DMAPacket;
	memset(GifTag, 0, sizeof(sceGifTag));
	DMAPacket++->ul64[0] = PS2_GIF_SET_TAG(m_imageSize >> 4, 
										GS_GIF_EOP_WITHFOLLOWINGPRIM, 
										GS_GIF_PRE_IGNOREPRIM, 
										0, 
										GIF_FLG_IMAGE, 
										0);
	//This end the tranfers of first cnt DMA tag, now build ref packets
	sps2uint32 PhyAddress;
	for (int pageNum = 0; pageNum < m_numMemBlocks - 1; ++pageNum) {
	
		PhyAddress = sps2GetPhysicalAddress( 
								(char *) m_imageData->pvStart + 4096 * pageNum, m_imageData);
		DMAPacket++->ul128 = PS2_DMA_SET_HEADER( 
								256, 0, DMA_ID_SOURCE_REF, 0, PhyAddress, 0);
	} 
	//Remember that last block can be partially filled with data
	int lastPageSize = (m_imageSize & 0x0FFF) >> 4;
	if (0 == lastPageSize) lastPageSize = 256;
	printf("lastPageSize = %d\n", lastPageSize);
	PhyAddress = sps2GetPhysicalAddress( 
								(char *) m_imageData->pvStart + 4096 * (m_numMemBlocks - 1), m_imageData);
	DMAPacket++->ul128 = PS2_DMA_SET_HEADER( 
							lastPageSize, 0, DMA_ID_SOURCE_REF, 0, PhyAddress, 0);
	//Now texture pixels are set, remember to set clut if necessary.
	//TODO, for now no CLUT is supported
	assert(m_pixelFormat == PSM_PSMCT32 || m_pixelFormat == PSM_PSMCT24);

	//Last GIF packet is used to access TEXFLUSH and stop data tranfer, this
	//also makes GS flushing data to be sure that next primitve that will be
	//rendered will correctly access the texture.
	DMAPacket++->ul128 = PS2_DMA_SET_HEADER(2, 0, DMA_ID_SOURCE_END, 0, 0, 0);
	GifTag = (sceGifTag *) DMAPacket;
	memset(GifTag, 0, sizeof(sceGifTag));
	//Remember to close GS packet, this is done setting EOP to 
	//GS_GIF_EOP_WITHOUTFOLLOWINGPRIM.
	DMAPacket++->ul64[0] = PS2_GIF_SET_TAG(1, 
								GS_GIF_EOP_WITHOUTFOLLOWINGPRIM, 
								GS_GIF_PRE_IGNOREPRIM, 
								0, 
								GIF_FLG_PACKED, 
								1);
	GifTag->REGS0 = GIF_REG_A_D;
	DMAPacket->ul64[0] = 0;
	DMAPacket++->ul64[1] = GIF_A_D_REG_TEXFLUSH;

	//Now the packet is to be send to GS, this transfer is different from 
	//the one used for past tutorial because we are doing source chain mode.
	//First of all we do not want any data to be transferred initially.
	//If 0 qword are to be transmitted MADR must not be set.
	*EE_D2_QWC	= 0;
	//This is the address of the first tag of the chain, this is memory
	//area passed by main program.
	*EE_D2_TADR = PS2_DMA_SET_TADR( 
		sps2GetPhysicalAddress(DMAMemory->pvStart, DMAMemory), 0);
	//Start the transfer.
	*EE_D2_CHCR = PS2_DMA_SET_CHCR(DMA_CHCR_DIR_TOMEMORY,
								DMA_CHCR_MOD_CHAIN,
								DMA_CHCR_ASP_NOADDRESS,
								DMA_CHCR_TTE_NOTRANSFER_DMATAG,
								DMA_CHCR_TIE_INTDISABLE,
								DMA_CHCR_STR_STARTTRANSFER,
								0);
	//wait for transfer to end.
	sps2WaitForDMA(2, m_SPS2Desc);
}

/** .
	 Since TEX0 register must be set by main application when this texture must
	 be used, this function build the structure of TEX0 register using texture
	 function passed as argument by the main program.
	 @param TextureFunction Texture function to use.
*/
sps2uint64 NablaPS2Utils::TextureManager::BuildTEX0(int TextureFunction) const {

	sps2GIFRegTEX0_n_t TEX0;
	TEX0.s.TBP0		= m_UploadAddress;
	TEX0.s.TBW		= m_width / 64;
	TEX0.s.PSM		= m_pixelFormat;
	TEX0.s.TW		= IntLog2(m_width);
	TEX0.s.TH		= IntLog2(m_height);
	TEX0.s.TCC		= GS_TEX0_TCC_RGB;
	TEX0.s.TFX		= TextureFunction;
	TEX0.s.CBP		= 0;
	TEX0.s.CPSM		= 0;
	TEX0.s.CSM		= 0;
	TEX0.s.CSA		= 0;
	TEX0.s.CLD		= 0;
	return		(sps2uint64) TEX0.i64;
}

/** .
	 Since TEX1 register must be set by main application when this texture must
	 be used but contains texture information, this function is used to build 
	 the structure of TEX1 register using filtering mode passed as argument 
	 by the main program.
	 @param FilteringMode It represent filtering mode for the texture, since
			  in this first implementatio mipmapping is not used same value
			  is used for minification as well for magnification.
*/
sps2uint64 NablaPS2Utils::TextureManager::BuildTEX1(int FilteringMode) const {

	sps2GIFRegTEX1_n_t TEX1;
	TEX1.s.LCM	= 0; //LODFORMULA but mipmapping is not used 
	TEX1.s.MXL	= 0;
	TEX1.s.MMAG = FilteringMode;
	TEX1.s.MMIN = FilteringMode;
	TEX1.s.MTBA = 0;
	TEX1.s.L		= 0;
	TEX1.s.K		= 0;
	return TEX1.i64;
}
/** .
	 This function update image data into texture manager.
*/
void NablaPS2Utils::TextureManager::UpdateImageData(const TextureFileHeader &TFH) {

	m_height			= TFH.Height;
	m_width			= TFH.Width;
	m_pixelFormat	= TFH.PixelFormat;
	m_swizzled		= TFH.Swizzled;
	m_imageSize		= TFH.DataSize; 
#ifdef DEBUG
	printf("Loaded Texture image");
	printf("Height\t\t = %d \n", m_height);
	printf("Width\t\t = %d \n", m_width);
	printf("PixelFormat\t = %d \n", m_pixelFormat);
	printf("Swizzled\t = %d \n", m_swizzled);
	printf("DataSize\t = %d \n", m_imageSize);
#endif
}

/** .


*/
int NablaPS2Utils::GetPixelSizeInBytes(PSM_t Format) {

	switch (Format) {
	
		case PSM_PSMCT32: 
		return 4;

		case PSM_PSMCT24:
		return 3;

		case PSM_PSMCT16: //Fall Through
		case PSM_PSMCT16S:
		return 2;
	
		case PSM_PSMT8:	//Fall Through
		case PSM_PSMT8H:
		case PSM_PSMT4:
		return 1;


		default:
			return 0;
	}	
}